#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import abc
import asyncio
import dataclasses
import ipaddress
import logging
import socket
import typing

import sipmessage
import websockets.asyncio.client
import websockets.asyncio.connection
import websockets.asyncio.server

from .utils import ANY_HOST, ANY_PORT, random_string

logger = logging.getLogger(__name__)

# Internal handlers.
ConnectHandler = typing.Callable[["TransportChannel"], None]
DisconnectHandler = typing.Callable[["TransportChannel", Exception | None], None]
MessageHandler = typing.Callable[["TransportChannel", sipmessage.Message], None]

# Public handlers.
RequestHandler = typing.Callable[
    [sipmessage.Request], typing.Coroutine[None, None, None]
]
ResponseHandler = typing.Callable[
    [sipmessage.Response], typing.Coroutine[None, None, None]
]

WEBSOCKET_SUBPROTOCOL = typing.cast(websockets.Subprotocol, "sip")


@dataclasses.dataclass
class TransportAddress:
    protocol: str
    'The protocol, e.g. `"udp"`, `"tcp"` or `"ws"`.'

    host: str
    'The transport host, e.g. `"127.0.0.1"`.'

    port: int
    "The transport port, e.g. `5060`."


class TransportChannel(abc.ABC):
    is_reliable: bool
    uri_transport: str
    remote_address: TransportAddress | None = None
    via_host: str
    via_port: int | None = None
    via_transport: str

    @abc.abstractmethod
    async def close(self) -> None: ...  # pragma: no cover

    @abc.abstractmethod
    async def send_message(
        self, message: sipmessage.Message, destination: TransportAddress
    ) -> None: ...  # pragma: no cover


def get_default_port(protocol: str) -> int:
    """
    Return the default port for the given protocol.
    """
    if protocol in ("ws", "wss"):
        return 0
    elif protocol == "tls":
        return 5061
    else:
        return 5060


async def get_transport_destination(message: sipmessage.Message) -> TransportAddress:
    """
    Determine the protocol, host and port to which a message should be sent.
    """
    protocol: str
    host: str
    port: int

    if isinstance(message, sipmessage.Request):
        # For requests, use the `Request-URI` or the first `Route` header.
        #
        # https://datatracker.ietf.org/doc/html/rfc3261#section-8.1.2
        route_set = message.route
        if route_set and "lr" in route_set[0].uri.parameters:
            # Loose routing, use the first `Route` header.
            uri = route_set[0].uri
        else:
            # Strict routing or empty route set, use the `Request-URI`.
            uri = message.uri

        # Check whether the target is a numeric IP address.
        try:
            ipaddress.ip_address(uri.host)
        except ValueError:
            is_ip_address = False
        else:
            is_ip_address = True

        # Select a transport protocol.
        #
        # As we do not support NAPTR lookups, we diverge from spec and
        # select either UDP or TLS depending on the the URI scheme.
        # This should theoretically only be done if the URI host is a
        # numeric IP address.
        #
        # https://datatracker.ietf.org/doc/html/rfc3263#section-4.1
        if "transport" in uri.parameters and uri.parameters["transport"]:
            # A transport was explicitly specified, use it.
            protocol = uri.parameters["transport"].lower()
        elif uri.scheme == "sips":
            # The URI scheme is "sips", use TLS.
            protocol = "tls"
        else:
            # Default to UDP.
            protocol = "udp"

        # Determine the IP address and port.
        #
        # https://datatracker.ietf.org/doc/html/rfc3263#section-4.2
        if is_ip_address or protocol == "ws":
            # The host is an IP address, use the specified port or the default
            # port for the transport protocol.
            #
            # We diverge from spec and also handle WebSockets here.
            host = uri.host
            port = uri.port or get_default_port(protocol)
        elif uri.port is not None:
            host = await asyncio.get_running_loop().run_in_executor(
                None, socket.gethostbyname, uri.host
            )
            port = uri.port
        else:
            raise NotImplementedError("SRV lookups are not yet supported")
    else:
        # For responses, use the top-most `Via` header.
        #
        # https://datatracker.ietf.org/doc/html/rfc3261#section-18.2.2
        via = message.via[0]

        # Select the transport protocol.
        protocol = via.transport.lower()

        # If the Via contains a "received" parameter, use it. Otherwise
        # use the "sent-by" value which is guaranteed to be an IP address.
        if "received" in via.parameters and via.parameters["received"]:
            host = via.parameters["received"]
        else:
            host = via.host

        # If the Via contains an "rport" parameter, use it. Otherwise use
        # the "sent-by value".
        if "rport" in via.parameters and via.parameters["rport"]:
            port = int(via.parameters["rport"])
        else:
            port = via.port or get_default_port(protocol)

    return TransportAddress(protocol=protocol, host=host, port=port)


def serialize_message(message: sipmessage.Message) -> bytes:
    """
    Serialise the message to bytes.

    This exists purely to keep type checks happy.
    """
    assert isinstance(message, (sipmessage.Request, sipmessage.Response))
    return bytes(message)


def set_transport_source(
    message: sipmessage.Message, channel: TransportChannel
) -> None:
    if isinstance(message, sipmessage.Request):
        # Rewrite the top-most Via with our address.
        vias = message.via
        if vias and vias[0].host == ANY_HOST:
            vias[0] = dataclasses.replace(
                vias[0],
                host=channel.via_host,
                port=channel.via_port,
                transport=channel.via_transport,
            )
            message.via = vias

    # Rewrite Contact headers with our address.
    contacts = []
    contacts_changed = False
    for contact in message.contact:
        if contact.uri.host == ANY_HOST:
            contact = dataclasses.replace(
                contact,
                uri=dataclasses.replace(
                    contact.uri,
                    host=channel.via_host,
                    port=channel.via_port,
                    parameters=contact.uri.parameters.replace(
                        transport=channel.uri_transport
                    ),
                ),
            )
            contacts_changed = True
        contacts.append(contact)
    if contacts_changed:
        message.contact = contacts


def update_request_via(
    request: sipmessage.Request, host: str, port: int, force: bool = False
) -> None:
    """
    Update the top-most `Via` header based on the IP address and port
    from which the request was received.
    """
    if request.via:
        via = request.via[0]
        via_changed = False

        # If the host portion of the `sent-by` in the top-most `Via` does
        # not match the IP address from which the request was received.
        #
        # https://datatracker.ietf.org/doc/html/rfc3261#section-18.2.1
        if force or via.host != host:
            via = dataclasses.replace(
                via, parameters=via.parameters.replace(received=host)
            )
            via_changed = True

        # If the top-most `Via` contains an `rport` parameter with no value, set
        # the `rport` parameter to the port from which the request was received.
        #
        # https://datatracker.ietf.org/doc/html/rfc3581#section-4
        if force or ("rport" in via.parameters and via.parameters["rport"] is None):
            via = dataclasses.replace(
                via, parameters=via.parameters.replace(rport=str(port))
            )
            via_changed = True

        if via_changed:
            request.via = [via] + request.via[1:]


class TransportLayer:
    """
    SIP transport layer which handles the transmission of requests and responses
    over network transports.

    See :rfc:`3261#section-18`.
    """

    def __init__(self) -> None:
        self._channels: set[TransportChannel] = set()
        self._tcp_servers: set[asyncio.Server] = set()
        self._ws_client = False
        self._ws_servers: set[websockets.asyncio.server.Server] = set()

        #: A coroutine which will be called whenever a request is received.
        self.request_handler: RequestHandler | None = None

        #: A coroutine which will be called whenever a response is received.
        self.response_handler: ResponseHandler | None = None

    async def close(self) -> None:
        """
        Close the transport layer.
        """
        for channel in set(self._channels):
            await channel.close()
        self._channels.clear()

        for tcp_server in self._tcp_servers:
            tcp_server.close()
        self._tcp_servers.clear()

        for ws_server in self._ws_servers:
            ws_server.close()
        self._ws_servers.clear()

    async def connect_websocket(self, uri: str) -> None:
        assert not self._channels, (
            "WebSocket client connection must be the only channel"
        )
        self._ws_client = True

        websocket = await websockets.asyncio.client.connect(
            uri, subprotocols=[WEBSOCKET_SUBPROTOCOL]
        )
        ready_event = asyncio.Event()
        asyncio.create_task(
            self._serve_websocket(
                is_secure=uri.startswith("wss:"),
                ready_event=ready_event,
                websocket=websocket,
            )
        )
        await ready_event.wait()

    async def listen(self, address: TransportAddress) -> None:
        """
        Start listening on the given transport address.
        """
        assert not self._ws_client, (
            "WebSocket client connection must be the only channel"
        )

        loop = asyncio.get_running_loop()
        if address.protocol == "tcp":
            tcp_server = await loop.create_server(
                lambda: TcpTransportChannel(
                    connect_handler=self._connect_handler,
                    disconnect_handler=self._disconnect_handler,
                    message_handler=self._message_handler,
                ),
                host=address.host,
                port=address.port,
            )
            self._tcp_servers.add(tcp_server)
        elif address.protocol == "ws":
            ws_server = await websockets.asyncio.server.serve(
                lambda websocket: self._serve_websocket(
                    is_secure=False, websocket=websocket
                ),
                address.host,
                address.port,
                subprotocols=[WEBSOCKET_SUBPROTOCOL],
            )
            self._ws_servers.add(ws_server)
        else:
            assert address.protocol == "udp"
            await loop.create_datagram_endpoint(
                lambda: UdpTransportChannel(
                    connect_handler=self._connect_handler,
                    disconnect_handler=self._disconnect_handler,
                    message_handler=self._message_handler,
                ),
                local_addr=(address.host, address.port),
            )

    async def send_message(self, message: sipmessage.Message) -> bool:
        """
        Send the given message and return whether the selected
        transport is reliable.
        """
        # Finalise the message.
        message.content_length = len(message.body)

        # Determine transport recipient.
        destination = await get_transport_destination(message)

        # Determine channel to use.
        channel = await self._acquire_channel(destination)

        # Rewrite references to our transport address.
        set_transport_source(message, channel)

        # Send the message and return whether a reliable transport was used.
        await channel.send_message(message, destination)
        return channel.is_reliable

    async def _acquire_channel(self, destination: TransportAddress) -> TransportChannel:
        """
        Find a suitable transport for the given destination.
        """
        for channel in self._channels:
            # A WebSocket client sends all its requests to the WebSocket server.
            if self._ws_client:
                return channel

            if (
                channel.remote_address is not None
                and channel.remote_address.protocol == destination.protocol
                and (channel.remote_address.host in (destination.host, ANY_HOST))
                and (channel.remote_address.port in (destination.port, ANY_PORT))
            ):
                return channel

        if destination.protocol == "tcp" and self._tcp_servers:
            loop = asyncio.get_running_loop()
            _transport, channel = await loop.create_connection(
                lambda: TcpTransportChannel(
                    connect_handler=self._connect_handler,
                    disconnect_handler=self._disconnect_handler,
                    message_handler=self._message_handler,
                ),
                host=destination.host,
                port=destination.port,
            )
            assert channel in self._channels
            return channel

        raise ConnectionError("No suitable transport found")

    async def _handle_message(self, message: sipmessage.Message) -> None:
        if isinstance(message, sipmessage.Request) and self.request_handler is not None:
            await self.request_handler(message)
        elif (
            isinstance(message, sipmessage.Response)
            and self.response_handler is not None
        ):
            # Check there is exactly one Via header, otherwise drop the response.
            #
            # https://datatracker.ietf.org/doc/html/rfc3261#section-8.1.3.3
            via_count = len(message.via)
            if via_count != 1:
                logger.warning(f"Expected exactly one Via header, got {via_count}")
                return
            await self.response_handler(message)

    def _connect_handler(self, channel: TransportChannel) -> None:
        self._channels.add(channel)

    def _disconnect_handler(
        self, channel: TransportChannel, exc: Exception | None
    ) -> None:
        self._channels.discard(channel)

    def _message_handler(
        self, channel: TransportChannel, message: sipmessage.Message
    ) -> None:
        asyncio.create_task(self._handle_message(message))

    async def _serve_websocket(
        self,
        *,
        is_secure: bool,
        websocket: websockets.asyncio.connection.Connection,
        ready_event: asyncio.Event | None = None,
    ) -> None:
        channel = WebsocketTransportChannel(
            is_secure=is_secure,
            message_handler=self._message_handler,
            websocket=websocket,
        )
        self._connect_handler(channel)
        if ready_event is not None:
            ready_event.set()
        try:
            async for data in websocket:
                if isinstance(data, bytes):
                    channel.message_received(data)
        finally:
            self._disconnect_handler(channel, None)


class TcpTransportChannel(TransportChannel, asyncio.Protocol):
    is_reliable = True
    uri_transport = "tcp"
    via_transport = "TCP"

    def __init__(
        self,
        *,
        connect_handler: ConnectHandler,
        disconnect_handler: DisconnectHandler,
        message_handler: MessageHandler,
    ) -> None:
        self._connect_handler = connect_handler
        self._disconnect_handler = disconnect_handler
        self._message_handler = message_handler
        self._transport: asyncio.Transport | None = None

    async def close(self) -> None:
        if self._transport is not None:
            self._transport.close()

    async def send_message(
        self, message: sipmessage.Message, destination: TransportAddress
    ) -> None:
        data = serialize_message(message)
        logger.info("=>\n\n" + data.decode())

        if self._transport is not None:
            self._transport.write(data)

    # Protocol

    def connection_lost(self, exc: Exception | None) -> None:
        self._disconnect_handler(self, exc)

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self._transport = typing.cast(asyncio.Transport, transport)

        sockname = transport.get_extra_info("sockname")
        self.via_host = sockname[0]
        self.via_port = sockname[1]

        peername = transport.get_extra_info("peername")
        self.remote_address = TransportAddress(
            protocol="tcp",
            host=peername[0],
            port=peername[1],
        )

        self._connect_handler(self)

    def data_received(self, data: bytes) -> None:
        logger.info("<=\n\n" + data.decode())

        try:
            message = sipmessage.Message.parse(data)
        except ValueError:
            # We cannot recover, close the socket.
            assert self._transport is not None
            self._transport.close()
            return
        else:
            if isinstance(message, sipmessage.Request):
                assert self.remote_address is not None
                update_request_via(
                    message, self.remote_address.host, self.remote_address.port
                )
            self._message_handler(self, message)


class UdpTransportChannel(TransportChannel, asyncio.DatagramProtocol):
    is_reliable = False
    remote_address = TransportAddress(
        protocol="udp",
        host=ANY_HOST,
        port=ANY_PORT,
    )
    uri_transport = "udp"
    via_transport = "UDP"

    def __init__(
        self,
        *,
        connect_handler: ConnectHandler,
        disconnect_handler: DisconnectHandler,
        message_handler: MessageHandler,
    ) -> None:
        self._connect_handler = connect_handler
        self._disconnect_handler = disconnect_handler
        self._message_handler = message_handler
        self._transport: asyncio.DatagramTransport | None = None

    async def close(self) -> None:
        if self._transport is not None:
            self._transport.close()

    async def send_message(
        self, message: sipmessage.Message, destination: TransportAddress
    ) -> None:
        data = serialize_message(message)
        logger.info("=>\n\n" + data.decode())

        if self._transport is not None:
            addr = (destination.host, destination.port)
            self._transport.sendto(data, addr)

    # DatagramProtocol

    def connection_lost(self, exc: Exception | None) -> None:
        self._disconnect_handler(self, exc)

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self._transport = typing.cast(asyncio.DatagramTransport, transport)

        sockname = transport.get_extra_info("sockname")
        self.via_host = sockname[0]
        self.via_port = sockname[1]

        self._connect_handler(self)

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        logger.info("<=\n\n" + data.decode())

        try:
            message = sipmessage.Message.parse(data)
        except ValueError:
            # Ignore the datagram.
            return
        else:
            if isinstance(message, sipmessage.Request):
                update_request_via(message, addr[0], addr[1])
            self._message_handler(self, message)


class WebsocketTransportChannel(TransportChannel):
    is_reliable = True
    uri_transport = "ws"

    def __init__(
        self,
        *,
        is_secure: bool,
        message_handler: MessageHandler,
        websocket: websockets.asyncio.connection.Connection,
    ) -> None:
        self._message_handler = message_handler
        self._websocket = websocket
        self.remote_address = TransportAddress(
            protocol="ws",
            host=websocket.remote_address[0],
            port=websocket.remote_address[1],
        )
        self.via_host = random_string(12) + ".invalid"
        self.via_port = None
        self.via_transport = "WSS" if is_secure else "WS"

    async def close(self) -> None:
        await self._websocket.close()

    async def send_message(
        self, message: sipmessage.Message, destination: TransportAddress
    ) -> None:
        data = serialize_message(message)
        logger.info("=>\n\n" + data.decode())

        await self._websocket.send(data)

    # WebSocket.

    def message_received(self, data: bytes) -> None:
        logger.info("<=\n\n" + data.decode())

        try:
            message = sipmessage.Message.parse(data)
        except ValueError:
            # Discard invalid message.
            return
        else:
            if isinstance(message, sipmessage.Request):
                assert self.remote_address is not None
                update_request_via(
                    message,
                    self.remote_address.host,
                    self.remote_address.port,
                    force=True,
                )
            self._message_handler(self, message)



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import secrets
import string

import sipmessage

ANY_HOST = "any"
ANY_PORT = 0
ANY_TRANSPORT = "ANY"

# https://datatracker.ietf.org/doc/html/rfc3261#section-21
SIP_STATUS_CODES = {
    100: "Trying",
    180: "Ringing",
    181: "Call Is Being Forwarded",
    182: "Queued",
    183: "Session Progress",
    200: "OK",
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Moved Temporarily",
    305: "Use Proxy",
    380: "Alternative Service",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    410: "Gone",
    413: "Request Entity Too Large",
    414: "Request-URI Too Long",
    415: "Unsupported Media Type",
    416: "Unsupported URI Scheme",
    420: "Bad Extension",
    421: "Extension Required",
    423: "Interval Too Brief",
    480: "Temporarily Unavailable",
    481: "Call/Transaction Does Not Exist",
    482: "Loop Detected",
    483: "Too Many Hops",
    484: "Address Incomplete",
    485: "Ambiguous",
    486: "Busy Here",
    487: "Request Terminated",
    488: "Not Acceptable Here",
    491: "Request Pending",
    493: "Undecipherable",
    500: "Server Internal Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Server Time-out",
    505: "Version Not Supported",
    513: "Message Too Large",
    600: "Busy Everywhere",
    603: "Decline",
    604: "Does Not Exist Anywhere",
    606: "Not Acceptable",
}


def create_ack(
    *,
    request: sipmessage.Request,
    response: sipmessage.Response,
) -> sipmessage.Request:
    """
    Create an ACK for the given request and response.

    See :rfc:`3261#section-17.1.1.3`.
    """
    ack = sipmessage.Request("ACK", request.uri)
    ack.via = [request.via[0]]
    ack.to_address = response.to_address
    ack.from_address = request.from_address
    ack.call_id = request.call_id
    ack.cseq = sipmessage.CSeq(sequence=request.cseq.sequence, method="ACK")

    ack.authorization = request.authorization
    ack.proxy_authorization = request.proxy_authorization
    ack.route = list(reversed(response.record_route))

    return ack


def create_contact(
    *,
    parameters: sipmessage.Parameters = sipmessage.Parameters(),
) -> sipmessage.Address:
    """
    Create a `Contact` header for a message.

    The host and transport will be filled in when the message is sent.
    """
    return sipmessage.Address(
        uri=sipmessage.URI(
            scheme="sip",
            host=ANY_HOST,
        ),
        parameters=parameters,
    )


def create_response(
    *,
    request: sipmessage.Request,
    code: int,
    phrase: str | None = None,
) -> sipmessage.Response:
    """
    Create a response for the given request.

    See :rfc:`3261#section-8.2.6`.
    """
    if phrase is None:
        phrase = SIP_STATUS_CODES[code]
    response = sipmessage.Response(code, phrase)
    response.via = request.via
    response.to_address = request.to_address
    response.from_address = request.from_address
    response.call_id = request.call_id
    response.cseq = request.cseq

    # Copy the Record-Route headers from the request if the response
    # establishes a dialog.
    #
    # https://datatracker.ietf.org/doc/html/rfc3261#section-12.1.1
    if response_establishes_dialog(request.method, code):
        response.record_route = request.record_route

    return response


def create_via() -> sipmessage.Via:
    """
    Create a `Via` header for a message.

    The host and transport will be filled in when the message is sent.
    """
    return sipmessage.Via(
        transport=ANY_TRANSPORT,
        host=ANY_HOST,
        parameters=sipmessage.Parameters(branch="z9hG4bK" + random_string(7)),
    )


def random_string(length: int) -> str:
    allchar = string.ascii_letters + string.digits
    return "".join(secrets.choice(allchar) for x in range(length))


def response_establishes_dialog(method: str, code: int) -> bool:
    return method == "INVITE" and code > 100 and code < 300



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import asyncio
import functools
from collections.abc import Callable, Coroutine
from typing import ParamSpec, TypeVar

import sipmessage

P = ParamSpec("P")
T = TypeVar("T")


def asynctest(
    coro: Callable[P, Coroutine[None, None, None]],
) -> Callable[P, None]:
    @functools.wraps(coro)
    def wrap(*args: P.args, **kwargs: P.kwargs) -> None:
        asyncio.run(coro(*args, **kwargs))

    return wrap


def lf2crlf(x: bytes) -> bytes:
    return x.replace(b"\n", b"\r\n")


def parse_request(data: bytes) -> sipmessage.Request:
    message = sipmessage.Message.parse(lf2crlf(data))
    assert isinstance(message, sipmessage.Request)
    return message


def parse_response(data: bytes) -> sipmessage.Response:
    message = sipmessage.Message.parse(lf2crlf(data))
    assert isinstance(message, sipmessage.Response)
    return message



import unittest

import sipmessage

from sipua.utils import create_ack, create_response

from .utils import lf2crlf, parse_request


class UtilsTest(unittest.TestCase):
    def assertMessage(
        self, message: sipmessage.Request | sipmessage.Response, data: bytes
    ) -> None:
        self.assertEqual(bytes(message), lf2crlf(data))

    def test_create_invite_response_and_ack(self) -> None:
        request = parse_request(b"""INVITE sip:+33233445566@127.0.0.1:5060 SIP/2.0
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
Max-Forwards: 70
To: sip:+33233445566@127.0.0.1:5060
From: sip:+33122334455@127.0.0.1:43248;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 INVITE
Content-Type: application/sdp
Content-Length: 235

v=0
o=- 1695503748 1695503748 IN IP4 127.0.0.1
s=-
c=IN IP4 127.0.0.1
t=0 0
m=audio 57299 RTP/AVP 0 8 101
a=rtpmap:0 PCMU/8000
a=rtpmap:8 PCMA/8000
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
a=silenceSupp:off - - - -
""")

        response = create_response(request=request, code=200)
        self.assertMessage(
            response,
            b"""SIP/2.0 200 OK
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
To: <sip:+33233445566@127.0.0.1:5060>
From: <sip:+33122334455@127.0.0.1:43248>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 INVITE

""",
        )

        ack = create_ack(request=request, response=response)
        self.assertMessage(
            ack,
            b"""ACK sip:+33233445566@127.0.0.1:5060 SIP/2.0
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
To: <sip:+33233445566@127.0.0.1:5060>
From: <sip:+33122334455@127.0.0.1:43248>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 ACK

""",
        )

    def test_create_invite_response_with_record_route(self) -> None:
        request = parse_request(b"""INVITE sip:callee@u2.domain.com SIP/2.0
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
Max-Forwards: 70
To: sip:callee@127.0.0.1:5060
From: sip:+caller@127.0.0.1:43248;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
Contact: sip:caller@u1.example.com
CSeq: 1 INVITE
Content-Type: application/sdp
Record-Route: <sip:p2.domain.com;lr>
Record-Route: <sip:p1.example.com;lr>

v=0
o=- 1695503748 1695503748 IN IP4 127.0.0.1
s=-
c=IN IP4 127.0.0.1
t=0 0
m=audio 57299 RTP/AVP 0 8 101
a=rtpmap:0 PCMU/8000
a=rtpmap:8 PCMA/8000
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
a=silenceSupp:off - - - -

""")
        response = create_response(request=request, code=200)
        self.assertMessage(
            response,
            b"""SIP/2.0 200 OK
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
To: <sip:callee@127.0.0.1:5060>
From: <sip:+caller@127.0.0.1:43248>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 INVITE
Record-Route: <sip:p2.domain.com;lr>
Record-Route: <sip:p1.example.com;lr>

""",
        )



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import asyncio
import enum
import logging
import typing

import sipmessage

from .transport import TransportLayer
from .utils import create_ack, create_response

logger = logging.getLogger(__name__)

T1 = 0.5
T2 = 4
T4 = 5

# Timer "D" is not defined relative to T1.
TIMER_D = 32

ClientTransactionKey = tuple[str, str]
ServerTransactionKey = tuple[str, str, str]

# Public handlers.
TransactionHandler = typing.Callable[
    ["ServerTransaction"],
    typing.Coroutine[None, None, None],
]


def get_client_transaction_key(message: sipmessage.Message) -> ClientTransactionKey:
    """
    Get the key used to match responses to client transactions.

    See :rfc:`3261#section-17.1.3`.
    """
    branch = message.via[0].parameters.get("branch")
    assert branch is not None, "Via branch must be defined for a transaction"

    if isinstance(message, sipmessage.Request):
        method = message.method
    else:
        method = message.cseq.method

    return (branch, method)


def get_server_transaction_key(request: sipmessage.Request) -> ServerTransactionKey:
    """
    Get the key used to match requests to server transactions.

    See :rfc:`3261#section-17.2.3`.
    """
    branch = request.via[0].parameters.get("branch")
    assert branch is not None, "Via branch must be defined for a transaction"

    sent_by = request.via[0].host
    if request.via[0].port is not None:
        sent_by += f":{request.via[0].port}"

    if request.method == "ACK":
        method = "INVITE"
    else:
        method = request.method

    return (branch, sent_by, method)


class TransactionState(enum.Enum):
    Trying = 0
    Calling = 1
    Proceeding = 2
    Completed = 3
    Confirmed = 4
    Terminated = 5


class Transaction:
    def __init__(
        self,
        *,
        request: sipmessage.Request,
        state: TransactionState,
        transaction_layer: "TransactionLayer",
    ) -> None:
        self._is_reliable = False
        self._request = request
        self._state = state
        self._transaction_layer = transaction_layer
        self._transport_layer = transaction_layer._transport_layer

    @property
    def request(self) -> sipmessage.Request:
        """
        The request which created the transaction.
        """
        return self._request

    def _log_info(self, msg: str, *args: object) -> None:
        logger.info(
            self.__class__.__name__ + " (" + self._request.method + ") " + msg, *args
        )

    def _set_state(self, state: TransactionState) -> None:
        if state != self._state:
            self._log_info("%s -> %s", self._state, state)
            self._state = state
            if state == TransactionState.Terminated:
                self._transaction_layer._transaction_terminated(self)

    def _start_timer(self, name: str, delay: float) -> None:
        cb = getattr(self, "_timer_%s_expired" % name.lower())
        loop = asyncio.get_event_loop()

        def wrap_cb() -> None:
            self._log_info("timer %s expired" % name)
            cb()

        self._log_info("timer %s started %.1f s" % (name, delay))
        loop.call_later(delay, wrap_cb)


class TransactionLayer:
    """
    SIP transaction layer, which handles matching responses to requests,
    retransmissions and timeouts.

    See :rfc:`3261#section-17`.
    """

    def __init__(self, transport_layer: TransportLayer) -> None:
        transport_layer.request_handler = self._receive_request
        transport_layer.response_handler = self._receive_response

        self._client_transactions: dict[ClientTransactionKey, ClientTransaction] = {}
        self._server_transactions: dict[ServerTransactionKey, ServerTransaction] = {}
        self._transport_layer = transport_layer

        #: A coroutine which will be called whenever a new transaction is received.
        self.transaction_handler: TransactionHandler | None = None

    async def request(self, request: sipmessage.Request) -> sipmessage.Response:
        """
        Send a request in a transaction and return the final response.
        """
        key = get_client_transaction_key(request)
        transaction: ClientInviteTransaction | ClientNonInviteTransaction
        if request.method == "INVITE":
            transaction = ClientInviteTransaction(
                request=request, transaction_layer=self
            )
        else:
            transaction = ClientNonInviteTransaction(
                request=request, transaction_layer=self
            )
        self._client_transactions[key] = transaction

        return await transaction.run()

    async def _receive_request(self, request: sipmessage.Request) -> None:
        # Try matching the request to a server transaction.
        key = get_server_transaction_key(request)
        transaction = self._server_transactions.get(key)
        if transaction is not None:
            await transaction._receive_request(request)
            return
        elif request.method == "ACK":
            # An un-matched ACK means the transaction has
            # already been terminated.
            return

        # Create a new server transaction.
        if request.method == "INVITE":
            transaction = ServerInviteTransaction(
                request=request, transaction_layer=self
            )
        else:
            transaction = ServerNonInviteTransaction(
                request=request, transaction_layer=self
            )
        self._server_transactions[key] = transaction
        await transaction._receive_request(request)

        # Notify the transaction user.
        if self.transaction_handler is not None:
            await self.transaction_handler(transaction)

    async def _receive_response(self, response: sipmessage.Response) -> None:
        # Try matching the response to a client transaction.
        key = get_client_transaction_key(response)
        transaction = self._client_transactions.get(key)
        if transaction is not None:
            await transaction.receive_response(response)

    def _transaction_terminated(self, transaction: Transaction) -> None:
        if isinstance(
            transaction, (ClientInviteTransaction, ClientNonInviteTransaction)
        ):
            server_key = get_client_transaction_key(transaction._request)
            self._client_transactions.pop(server_key)
        elif isinstance(
            transaction, (ServerInviteTransaction, ServerNonInviteTransaction)
        ):
            client_key = get_server_transaction_key(transaction._request)
            self._server_transactions.pop(client_key)


class ClientInviteTransaction(Transaction):
    """
    An INVITE client transaction.

    See :rfc:`3261#section-17.1.2`.
    """

    def __init__(
        self, *, request: sipmessage.Request, transaction_layer: TransactionLayer
    ) -> None:
        assert request.method == "INVITE", "Invalid method"
        super().__init__(
            request=request,
            state=TransactionState.Calling,
            transaction_layer=transaction_layer,
        )
        self._future: asyncio.Future[sipmessage.Response] = asyncio.Future()

    async def receive_response(self, response: sipmessage.Response) -> None:
        """
        Handle a response from the transport layer.
        """
        if self._state in (TransactionState.Calling, TransactionState.Proceeding):
            if response.code >= 100 and response.code < 200:
                self._set_state(TransactionState.Proceeding)
            elif response.code >= 200 and response.code < 300:
                # Send ACK.
                ack = create_ack(request=self._request, response=response)
                await self._transport_layer.send_message(ack)

                self._set_state(TransactionState.Terminated)
                self._future.set_result(response)
            elif response.code >= 300 and response.code < 700:
                # Send ACK.
                ack = create_ack(request=self._request, response=response)
                await self._transport_layer.send_message(ack)

                self._set_state(TransactionState.Completed)
                self._future.set_result(response)

                self._start_timer("D", not self._is_reliable and TIMER_D or 0)

    async def run(self) -> sipmessage.Response:
        """
        Send the request and return the response. If no response is received, a
        :class:`TimeoutError` is raised.
        """

        self._is_reliable = await self._transport_layer.send_message(self._request)
        self._start_timer("B", 64 * T1)

        # For unreliable transports, schedule retransmission.
        if not self._is_reliable:
            self.__timer_a_value = T1
            self._start_timer("A", self.__timer_a_value)

        return await self._future

    def _timer_a_expired(self) -> None:
        """
        Re-transmit the request while the transaction is in the `Calling` state.

        This is only used for unreliable transports.
        """
        if self._state == TransactionState.Calling:
            asyncio.ensure_future(self._transport_layer.send_message(self._request))
            self.__timer_a_value *= 2
            self._start_timer("A", self.__timer_a_value)

    def _timer_b_expired(self) -> None:
        """
        Terminate the transaction if it is still in the `Calling` or `Proceeding`
        state.
        """
        if self._state in (TransactionState.Calling, TransactionState.Proceeding):
            self._set_state(TransactionState.Terminated)
            self._future.set_exception(TimeoutError)

    def _timer_d_expired(self) -> None:
        """
        Terminate the transaction after it reaches the `Completed` state.
        """
        if self._state == TransactionState.Completed:
            self._set_state(TransactionState.Terminated)


class ClientNonInviteTransaction(Transaction):
    """
    A non-INVITE client transaction.

    See :rfc:`3261#section-17.1.2`.
    """

    def __init__(
        self, *, request: sipmessage.Request, transaction_layer: TransactionLayer
    ) -> None:
        assert request.method != "INVITE", "Invalid method"
        super().__init__(
            request=request,
            state=TransactionState.Trying,
            transaction_layer=transaction_layer,
        )
        self._future: asyncio.Future[sipmessage.Response] = asyncio.Future()

    async def receive_response(self, response: sipmessage.Response) -> None:
        """
        Handle a response from the transport layer.
        """
        if self._state in (TransactionState.Trying, TransactionState.Proceeding):
            if response.code >= 100 and response.code < 200:
                self._set_state(TransactionState.Proceeding)
            elif response.code >= 200 and response.code < 700:
                self._set_state(TransactionState.Completed)
                self._future.set_result(response)

                self._start_timer("K", not self._is_reliable and T4 or 0)

    async def run(self) -> sipmessage.Response:
        """
        Send the request and return the response. If no response is received, a
        :class:`TimeoutError` is raised.
        """
        self._is_reliable = await self._transport_layer.send_message(self._request)
        self._start_timer("F", 64 * T1)

        # For unreliable transports, schedule retransmission.
        if not self._is_reliable:
            self.__timer_e_value = T1
            self._start_timer("E", self.__timer_e_value)

        return await self._future

    def _timer_e_expired(self) -> None:
        """
        Re-transmit the request while the transaction is in the `Trying` state.

        This is only used for unreliable transports.
        """
        if self._state == TransactionState.Trying:
            asyncio.ensure_future(self._transport_layer.send_message(self._request))
            self.__timer_e_value = min(2 * self.__timer_e_value, T2)
            self._start_timer("E", self.__timer_e_value)

    def _timer_f_expired(self) -> None:
        """
        Terminate the transaction if it is still in the `Trying` or `Proceeding`
        state.
        """
        if self._state in (TransactionState.Trying, TransactionState.Proceeding):
            self._set_state(TransactionState.Terminated)
            self._future.set_exception(TimeoutError)

    def _timer_k_expired(self) -> None:
        """
        Terminate the transaction after it reaches the `Completed` state.
        """
        if self._state == TransactionState.Completed:
            self._set_state(TransactionState.Terminated)


class ServerInviteTransaction(Transaction):
    """
    An INVITE server transaction.

    See :rfc:`3261#section-17.2.1`.
    """

    def __init__(
        self, *, request: sipmessage.Request, transaction_layer: TransactionLayer
    ) -> None:
        assert request.method == "INVITE", "Invalid method"
        super().__init__(
            request=request,
            state=TransactionState.Proceeding,
            transaction_layer=transaction_layer,
        )

    async def send_response(self, response: sipmessage.Response) -> None:
        """
        Send the given response.
        """
        if self._state == TransactionState.Proceeding:
            if response.code >= 100 and response.code < 200:
                self._is_reliable = await self._transport_layer.send_message(response)
            elif response.code >= 200 and response.code < 300:
                self._is_reliable = await self._transport_layer.send_message(response)
                self._set_state(TransactionState.Terminated)
            elif response.code >= 300 and response.code < 700:
                self._is_reliable = await self._transport_layer.send_message(response)
                self._set_state(TransactionState.Completed)
                self._start_timer("H", 64 * T1)

                # For unreliable transports, schedule retransmission.
                if not self._is_reliable:
                    self.response = response
                    self.__timer_g_value = T1
                    self._start_timer("G", self.__timer_g_value)

    async def _receive_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request from the transport layer.
        """
        if self._state == TransactionState.Proceeding and request.method == "INVITE":
            response = create_response(request=request, code=100)
            await self.send_response(response)
        elif self._state == TransactionState.Completed and request.method == "ACK":
            self._set_state(TransactionState.Confirmed)

            self._start_timer("I", not self._is_reliable and T4 or 0)

    def _timer_g_expired(self) -> None:
        """
        Re-transmit a 300-699 response while the transaction is in the `Completed`
        state.

        This is only used for unreliable transports.
        """
        if self._state == TransactionState.Completed:
            asyncio.ensure_future(self._transport_layer.send_message(self.response))
            self.__timer_g_value = min(2 * self.__timer_g_value, T2)
            self._start_timer("G", self.__timer_g_value)

    def _timer_h_expired(self) -> None:
        """
        Terminate the transaction after it reaches the `Completed` state.
        """
        if self._state == TransactionState.Completed:
            self._set_state(TransactionState.Terminated)

    def _timer_i_expired(self) -> None:
        """
        Terminate the transaction after it reaches the `Confirmed` state.
        """
        if self._state == TransactionState.Confirmed:
            self._set_state(TransactionState.Terminated)


class ServerNonInviteTransaction(Transaction):
    """
    A non-INVITE server transaction.

    See :rfc:`3261#section-17.2.2`.
    """

    def __init__(
        self, *, request: sipmessage.Request, transaction_layer: TransactionLayer
    ) -> None:
        assert request.method != "INVITE", "Invalid method"
        super().__init__(
            request=request,
            state=TransactionState.Trying,
            transaction_layer=transaction_layer,
        )

    async def send_response(self, response: sipmessage.Response) -> None:
        """
        Send the given response.
        """
        if self._state in (TransactionState.Trying, TransactionState.Proceeding):
            if response.code >= 100 and response.code < 200:
                self._is_reliable = await self._transport_layer.send_message(response)
                self._set_state(TransactionState.Proceeding)
            elif response.code >= 200 and response.code < 700:
                self._is_reliable = await self._transport_layer.send_message(response)
                self._set_state(TransactionState.Completed)

                self._start_timer("J", not self._is_reliable and 64 * T1 or 0)

    async def _receive_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request from the transport layer.
        """
        pass

    def _timer_j_expired(self) -> None:
        """
        Terminate the transaction after it reaches the `Completed` state.
        """
        if self._state == TransactionState.Completed:
            self._set_state(TransactionState.Terminated)


ClientTransaction = ClientInviteTransaction | ClientNonInviteTransaction
ServerTransaction = ServerInviteTransaction | ServerNonInviteTransaction



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import asyncio
import contextlib
import typing
import unittest
from collections.abc import AsyncGenerator, Callable
from unittest.mock import patch

import sipmessage

import sipua.transaction as sip_transaction
from sipua.transaction import (
    ClientInviteTransaction,
    ClientNonInviteTransaction,
    ServerInviteTransaction,
    ServerNonInviteTransaction,
    ServerTransaction,
    TransactionLayer,
    TransactionState,
    get_client_transaction_key,
)
from sipua.transport import TransportAddress, TransportChannel, TransportLayer
from sipua.utils import ANY_HOST, ANY_PORT, create_response

from .utils import asynctest

P = typing.ParamSpec("P")
T = typing.TypeVar("T")

FROM_ADDRESS = sipmessage.Address(
    uri=sipmessage.URI(scheme="sip", user="alice", host="127.0.0.1", port=5061)
)
TO_ADDRESS = sipmessage.Address(
    uri=sipmessage.URI(scheme="sip", user="bob", host="127.0.0.1", port=5062)
)


class DummyTransportChannel(TransportChannel):
    is_reliable = False
    remote_address = TransportAddress(
        protocol="udp",
        host=ANY_HOST,
        port=ANY_PORT,
    )
    uri_transport = "udp"
    via_transport = "UDP"

    def __init__(self, outbox: list[sipmessage.Message]) -> None:
        self._outbox = outbox

    async def close(self) -> None:
        pass

    async def send_message(
        self, message: sipmessage.Message, destination: TransportAddress
    ) -> None:
        self._outbox.append(message)


def create_request(method: str, sequence: int = 1) -> sipmessage.Request:
    request = sipmessage.Request(method, TO_ADDRESS.uri)
    request.call_id = "12345"
    request.cseq = sipmessage.CSeq(sequence=sequence, method=method)
    request.from_address = FROM_ADDRESS
    request.to_address = TO_ADDRESS
    request.via = [
        sipmessage.Via(
            transport="UDP",
            host="127.0.0.1",
            port=5061,
            parameters=sipmessage.Parameters(branch="z9hG4bK1e5b2b763d"),
        )
    ]
    return request


def shorten_timers(func: Callable[P, T]) -> Callable[P, T]:
    """
    Reduce transaction timers by 100.
    """
    return patch("sipua.transaction.T1", sip_transaction.T1 / 100)(
        patch("sipua.transaction.T2", sip_transaction.T2 / 100)(
            patch("sipua.transaction.T4", sip_transaction.T4 / 100)(
                patch("sipua.transaction.TIMER_D", sip_transaction.TIMER_D / 100)(func)
            )
        )
    )


class BaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.message_outbox: list[sipmessage.Message] = []
        self.transport = TransportLayer()
        self.transport._channels.add(DummyTransportChannel(self.message_outbox))
        self.transaction_layer = TransactionLayer(transport_layer=self.transport)

    def assertSentRequests(self, methods: list[str]) -> None:
        sent_methods = [
            m.method for m in self.message_outbox if isinstance(m, sipmessage.Request)
        ]
        self.assertEqual(sent_methods[0 : len(methods)], methods)

    def assertSentResponses(self, codes: list[int]) -> None:
        sent_codes = [
            m.code for m in self.message_outbox if isinstance(m, sipmessage.Response)
        ]
        self.assertEqual(sent_codes[0 : len(codes)], codes)

    async def receive_messages(self, *messages: sipmessage.Message) -> None:
        for message in messages:
            if isinstance(message, sipmessage.Request):
                await self.transaction_layer._receive_request(message)
            elif isinstance(message, sipmessage.Response):
                await self.transaction_layer._receive_response(message)


class ClientInviteTransactionTest(BaseTestCase):
    def create_transaction(
        self, request: sipmessage.Request
    ) -> ClientInviteTransaction:
        transaction = ClientInviteTransaction(
            request=request, transaction_layer=self.transaction_layer
        )
        self.assertEqual(transaction._state, TransactionState.Calling)
        key = get_client_transaction_key(request)
        self.transaction_layer._client_transactions[key] = transaction
        return transaction

    @asynctest
    async def test_response_200(self) -> None:
        request = create_request("INVITE")
        response_100 = create_response(request=request, code=100)
        response_200 = create_response(request=request, code=200)

        transaction = self.create_transaction(request)

        result = await asyncio.gather(
            transaction.run(),
            self.receive_messages(response_100, response_200),
        )
        self.assertEqual(result, [response_200, None])
        self.assertEqual(transaction._state, TransactionState.Terminated)
        self.assertSentRequests(["INVITE", "ACK"])

    @asynctest
    @shorten_timers
    async def test_response_404(self) -> None:
        request = create_request("INVITE")
        response_404 = create_response(request=request, code=404)

        transaction = self.create_transaction(request)

        result = await asyncio.gather(
            transaction.run(),
            self.receive_messages(response_404),
        )
        self.assertEqual(result, [response_404, None])
        self.assertEqual(transaction._state, TransactionState.Completed)
        self.assertSentRequests(["INVITE", "ACK"])

        # Allow timer D to expire.
        await asyncio.sleep(sip_transaction.TIMER_D)
        self.assertEqual(transaction._state, TransactionState.Terminated)

    @asynctest
    @shorten_timers
    async def test_timeout(self) -> None:
        request = create_request("INVITE")

        transaction = self.create_transaction(request)

        with self.assertRaises(TimeoutError):
            await transaction.run()
        self.assertEqual(transaction._state, TransactionState.Terminated)
        self.assertSentRequests(
            ["INVITE", "INVITE", "INVITE", "INVITE", "INVITE", "INVITE", "INVITE"]
        )


class ClientNonInviteTransactionTest(BaseTestCase):
    def create_transaction(
        self, request: sipmessage.Request
    ) -> ClientNonInviteTransaction:
        transaction = ClientNonInviteTransaction(
            request=request, transaction_layer=self.transaction_layer
        )
        self.assertEqual(transaction._state, TransactionState.Trying)
        key = get_client_transaction_key(request)
        self.transaction_layer._client_transactions[key] = transaction
        return transaction

    @asynctest
    async def test_response_200(self) -> None:
        request = create_request("REGISTER")
        response_100 = create_response(request=request, code=100)
        response_200 = create_response(request=request, code=200)

        transaction = self.create_transaction(request)

        result = await asyncio.gather(
            transaction.run(),
            self.receive_messages(response_100, response_200),
        )
        self.assertEqual(result, [response_200, None])
        self.assertEqual(transaction._state, TransactionState.Completed)
        self.assertSentRequests(["REGISTER"])

    @asynctest
    @shorten_timers
    async def test_response_404(self) -> None:
        request = create_request("REGISTER")
        response_404 = create_response(request=request, code=404)

        transaction = self.create_transaction(request)

        result = await asyncio.gather(
            transaction.run(),
            self.receive_messages(response_404),
        )
        self.assertEqual(result, [response_404, None])
        self.assertEqual(transaction._state, TransactionState.Completed)
        self.assertSentRequests(["REGISTER"])

        # Allow timer K to expire.
        await asyncio.sleep(sip_transaction.T4)
        self.assertEqual(transaction._state, TransactionState.Terminated)

    @asynctest
    @shorten_timers
    async def test_timeout(self) -> None:
        request = create_request("REGISTER")

        transaction = self.create_transaction(request)

        with self.assertRaises(TimeoutError):
            await transaction.run()
        self.assertEqual(transaction._state, TransactionState.Terminated)
        self.assertSentRequests(
            [
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
                "REGISTER",
            ]
        )


class ServerInviteTransactionTest(BaseTestCase):
    async def create_transaction(
        self, request: sipmessage.Request
    ) -> ServerInviteTransaction:
        await self.transaction_layer._receive_request(request)

        transactions = list(self.transaction_layer._server_transactions.values())
        self.assertEqual(len(transactions), 1)

        transaction = transactions[0]
        assert isinstance(transaction, ServerInviteTransaction)
        self.assertEqual(transaction._state, TransactionState.Proceeding)

        return transaction

    @asynctest
    @shorten_timers
    async def test_response_200(self) -> None:
        request = create_request("INVITE")
        transaction = await self.create_transaction(request)

        response = create_response(request=request, code=200)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        ack = create_request("ACK", sequence=2)
        await self.receive_messages(ack)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        self.assertSentResponses([100, 200])

    @asynctest
    @shorten_timers
    async def test_response_404(self) -> None:
        request = create_request("INVITE")
        transaction = await self.create_transaction(request)

        response = create_response(request=request, code=404)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Completed)

        ack = create_request("ACK", sequence=2)
        await self.receive_messages(ack)
        self.assertEqual(transaction._state, TransactionState.Confirmed)

        # Allow timer I to expire.
        await asyncio.sleep(sip_transaction.T4)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        self.assertSentResponses([100, 404])

    @asynctest
    @shorten_timers
    async def test_response_404_timeout(self) -> None:
        request = create_request("INVITE")
        transaction = await self.create_transaction(request)

        response = create_response(request=request, code=404)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Completed)

        # Allow timer G and H to expire.
        await asyncio.sleep(sip_transaction.T1 * 64)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        self.assertSentResponses([100, 404, 404, 404, 404, 404, 404, 404, 404, 404])


class ServerNonInviteTransactionTest(BaseTestCase):
    async def create_transaction(
        self, request: sipmessage.Request
    ) -> ServerNonInviteTransaction:
        await self.transaction_layer._receive_request(request)

        transactions = list(self.transaction_layer._server_transactions.values())
        self.assertEqual(len(transactions), 1)

        transaction = transactions[0]
        assert isinstance(transaction, ServerNonInviteTransaction)
        self.assertEqual(transaction._state, TransactionState.Trying)

        return transaction

    @asynctest
    @shorten_timers
    async def test_response_100_then_200(self) -> None:
        request = create_request("REGISTER")
        transaction = await self.create_transaction(request)

        response = create_response(request=request, code=100)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Proceeding)

        response = create_response(request=request, code=200)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Completed)

        # Allow timer J to expire.
        await asyncio.sleep(sip_transaction.T1 * 64)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        self.assertSentResponses([100, 200])

    @asynctest
    @shorten_timers
    async def test_response_200(self) -> None:
        request = create_request("REGISTER")
        transaction = await self.create_transaction(request)

        response = create_response(request=request, code=200)
        await transaction.send_response(response)
        self.assertEqual(transaction._state, TransactionState.Completed)

        # Allow timer J to expire.
        await asyncio.sleep(sip_transaction.T1 * 64)
        self.assertEqual(transaction._state, TransactionState.Terminated)

        self.assertSentResponses([200])


class EndToEndTest(unittest.TestCase):
    @contextlib.asynccontextmanager
    async def client_and_server(
        self,
    ) -> AsyncGenerator[tuple[TransactionLayer, TransactionLayer], None]:
        client_transport = TransportLayer()
        client_transaction = TransactionLayer(client_transport)

        server_transport = TransportLayer()
        server_transaction = TransactionLayer(server_transport)

        try:
            await client_transport.listen(TransportAddress("udp", "127.0.0.1", 5061))
            await server_transport.listen(TransportAddress("udp", "127.0.0.1", 5062))
            yield (client_transaction, server_transaction)
        finally:
            await client_transport.close()
            await server_transport.close()

    @asynctest
    async def test_invite(self) -> None:
        async def reply_to_invite(transaction: ServerTransaction) -> None:
            if transaction.request.method == "INVITE":
                response = create_response(request=transaction.request, code=200)
                await transaction.send_response(response)

        async with self.client_and_server() as (client_transaction, server_transaction):
            server_transaction.transaction_handler = reply_to_invite

            request = create_request("INVITE")
            response = await client_transaction.request(request)
            self.assertEqual(response.code, 200)

    @asynctest
    async def test_register(self) -> None:
        async def reply_to_register(transaction: ServerTransaction) -> None:
            if transaction.request.method == "REGISTER":
                response = create_response(request=transaction.request, code=200)
                await transaction.send_response(response)

        async with self.client_and_server() as (client_transaction, server_transaction):
            server_transaction.transaction_handler = reply_to_register

            request = create_request("REGISTER")
            response = await client_transaction.request(request)
            self.assertEqual(response.code, 200)



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import importlib.metadata

from .transaction import (
    ServerInviteTransaction,
    ServerNonInviteTransaction,
    ServerTransaction,
    TransactionLayer,
)
from .transport import TransportAddress, TransportLayer
from .utils import create_ack, create_response

__all__ = [
    "create_ack",
    "create_response",
    "ServerInviteTransaction",
    "ServerNonInviteTransaction",
    "ServerTransaction",
    "TransactionLayer",
    "TransportAddress",
    "TransportLayer",
]
__version__ = importlib.metadata.version("sipua")
